@@ -63,9 +63,13 @@ WEBSOCKET_PORT = 5001
 #
 # These must match the `constants in the client <GetResultType JS>`.
 class GetResultType(Enum):
-    html = 0
+    # A URL indicating that new rendered content is available.
+    url = 0
+    # A build output message.
     build = 1
+    # Errors from the build.
     errors = 2
+    # A command, such as ``shutdown```.
     command = 3
 
 
@@ -578,22 +582,23 @@ async def convert_file(text: str, file_path: str, cs: ClientState) -> None:
 
     if asyncio.iscoroutinefunction(converter):
         # Coroutines get the queue, so they can report progress during the build.
-        html_string, err_string = await converter(
+        html_string_or_file_path, err_string = await converter(
             text, file_path, tool_or_project_path, cs.q
         )
     else:
         assert tool_or_project_path is None
-        html_string, err_string = converter(text, file_path)
+        html_string_or_file_path, err_string = converter(text, file_path)
 
     # Update the client's state, now that the rendering is complete.
     cs._editor_text = text
-    # For projects, the rendered HTML is already on disk.
     if is_project:
-        cs._file_path = html_string
+        # For projects, the rendered HTML is already on disk; a path to this rendered file is returned.
+        cs._file_path = html_string_or_file_path
         cs._html = None
     else:
+        # Otherwise, the rendered HTML is returned as a string and can be directly used. Provide a path to the source file which was just rendered.
         cs._file_path = file_path
-        cs._html = html_string
+        cs._html = html_string_or_file_path
 
     # Send any errors. An empty error string will clear any errors from a previous build, and should still be sent.
     await cs.q.put(GetResultReturn(GetResultType.errors, err_string))
@@ -603,7 +608,7 @@ async def convert_file(text: str, file_path: str, cs: ClientState) -> None:
     # For Windows, make the path contain forward slashes.
     uri = path_to_uri(cs._file_path)
     # Encode this, for Windows paths which contain a colon (or unusual Linux paths).
-    await cs.q.put(GetResultReturn(GetResultType.html, urllib.parse.quote(uri)))
+    await cs.q.put(GetResultReturn(GetResultType.url, urllib.parse.quote(uri)))
 
 
 # RenderManager / render thread
@@ -702,7 +707,10 @@ class RenderManager:
         cs = self.get_client_state(id)
         return cast(ClientState, cs).q if cs else None
 
-    # Given a URL, see if it matches with the latest render; if so, return the resulting HTML. If there's no match to the URL or the ID doesn't exist, return False. Note that the "HTML" can be None, meaning the render was stored to disk and the URL is a path to the rendered file.
+    # Return the results of rendering the provided URL:
+    #
+    # - If the URL matches with the latest render, return the resulting HTML for a non-project render. Return ``None`` for a project render, indicating that the render was stored to disk and the URL is a path to the rendered file.
+    # - If there's no match to the URL or the ID doesn't exist, return False. Note that the "HTML" can be None, meaning
     def get_render_results(self, id: int, url_path: str) -> Union[None, str, bool]:
         cs = self.get_client_state(id)
         return (
@@ -717,18 +725,24 @@ class RenderManager:
     ):
         # First, read this client's ID.
         try:
-            id = json.loads(await websocket.recv())
+            data = await websocket.recv()
         except websockets.exceptions.WebSocketException:
             # Give up if there's a websocket error.
             return
 
         # Find the queue for this client.
+        try:
+            id = json.loads(data)
+        except json.decoder.JSONDecodeError:
+            id = f"<invalid id {repr(data)}>"
         q = self.get_queue(id)
         if not q:
             try:
                 await websocket.send(
-                    GetResultReturn(
-                        GetResultType.command, f"error: unknown client {id}."
+                    json.dumps(
+                        GetResultReturn(
+                            GetResultType.command, f"error: unknown client {id}."
+                        )
                     )
                 )
             except websockets.exceptions.WebSocketException:
@@ -825,6 +839,8 @@ class RenderManager:
         )
         # Let the user know that the server is now ready -- this is the last piece of it to start.
         print("Ready.")
+        # For testing, flush this since test code waits for it before running the rest of a test.
+        sys.stdout.flush()
         await asyncio.gather(*[self._worker(i) for i in range(num_workers)])
 
     # Process items in the render queue.
